"""
This file contains global variables used by SUS.
"""

CLIENT_ENC_NONCE = b"cliqencc"
"Client encryption nonce"

SERVER_ENC_NONCE = b"cliqencs"
"Server encryption nonce"

CLIENT_MAC_NONCE = b"cliqmacc"
"Client MAC nonce"

SERVER_MAC_NONCE = b"cliqmacs"
"Server MAC nonce"
